clc;clear;
tic

param.a1 = -5;
param.a2 = -3;
param.h1 = 5;
param.h2 = 3;
param.c = -0.2;
param.d = 20;

% Objective function
F       = @(X,Y) param.a1*(X-param.h1).^3 +param.a2*(Y-param.h2).^3 + param.c*X.*Y + param.d;
F_X     = @(X,Y) 3*param.a1*(X-param.h1).^2 + param.c.*Y;
F_Y     = @(X,Y) 3*param.a2*(Y-param.h1).^2 + param.c.*X;
F_XY    = @(X,Y) param.c+0*X+0*Y;
F_XX    = @(X,Y) 6*param.a1*(X-param.h1);
F_YY    = @(X,Y) 6*param.a2*(Y-param.h2);

xplotmin    = -1;
xplotmax    = 10;
glob.Nx     = 50;
xgrid0      = nodeunif(glob.Nx,xplotmin,xplotmax);
ygrid0      = nodeunif(glob.Nx,xplotmin,xplotmax);
fspace      = fundef({'spli',xgrid0,0,3},{'spli',ygrid0,0,3});
sgrid       = funnode(fspace);
s           = gridmake(sgrid);
xgrid       = s(s(:,2)==s(1,2),1);
ygrid       = s(s(:,1)==s(1,1),2);

Fgrid       = F(s(:,1),s(:,2)); 
c           = funfitxy(fspace,s,Fgrid);
glob.xgrid  = xgrid;
glob.xgrid0 = xgrid0;
glob.ygrid  = ygrid;
glob.ygrid0 = ygrid0;

xplot      = nodeunif(10,xplotmin,xplotmax);
yplot      = nodeunif(10,xplotmin,xplotmax);
xyplot     = gridmake(xplot,yplot);

[D1_X,D2_X]         = SM_splidop(xgrid0);
[B0_X,B1_X,B2_X]    = SM_splibas(xgrid0,xyplot(:,1),D1_X,D2_X,3);     

[D1_Y,D2_Y]         = SM_splidop(ygrid0);
[B0_Y,B1_Y,B2_Y]    = SM_splibas(ygrid0,xyplot(:,2),D1_Y,D2_Y,3);     

B0_X_B0_Y           = dprod(B0_Y,B0_X);
B1_X_B0_Y           = dprod(B0_Y,B1_X);
B0_X_B1_Y           = dprod(B1_Y,B0_X);
B1_X_B1_Y           = dprod(B1_Y,B1_X);
B2_X_B0_Y           = dprod(B0_Y,B2_X);
B0_X_B2_Y           = dprod(B2_Y,B0_X);

Xmin = 1;
Xmax = 9;
Ymin = 1;
Ymax = 9;

[D1_X,D2_X]         = SM_splidop(xgrid0);
[D1_Y,D2_Y]         = SM_splidop(ygrid0);
glob.D1_X           = D1_X;
glob.D2_X           = D2_X;
glob.D1_Y           = D1_Y;
glob.D2_Y           = D2_Y;
h                   = 1e-6;

%% Use fsolve
tic 
mstarfs   = fsolve(@(m) foc_con_2D(m,h,Xmin,Xmax,Ymin,Ymax,c,param,glob),[.5;.5]);
[FOCfs,~,Xfs,Yfs,L1_Xfs,L2_Xfs,L1_Yfs,L2_Yfs] = foc_con_2D(mstarfs,h,Xmin,Xmax,Ymin,Ymax,c,param,glob);
tocfsolve = toc; 

%% Using Newton-solver
tic

% 1. Compute derivative operator matrices
[D1_X,D2_X]         = SM_splidop(xgrid0);
[D1_Y,D2_Y]         = SM_splidop(ygrid0);
glob.D1_X           = D1_X;
glob.D2_X           = D2_X;
glob.D1_Y           = D1_X;
glob.D2_Y           = D2_Y;

h                       = 1e-6;
m0                      = [.3;.8];

% Solution 
[g0,D0,FOCs,Fs,X0,Y0]   = foc_con_2D(m0,h,Xmin,Xmax,Ymin,Ymax,c,param,glob); 
max_iter = 1000;
for iter0 = 1:max_iter
    %______________________________________________________________________
    % 1. Update
    upd         = - D0\g0;
    m1          = m0 + upd;
    %______________________________________________________________________
    % 2. Backstep (optional)
%     g1          = foc_con_2D(m1,F,h,Xmin,Xmax,Ymin,Ymax);
%     n           = 1;
%     while max(abs(g1))>max(abs(g0)) && n<10;
%         upd     = (1/2) * upd;
%         m1      = m0 - upd;
%         g1      = foc_con_2D(m1,h,Xmin,Xmax,Ymin,Ymax,c,param,glob);
%         n       = n+1;
%     end
    %______________________________________________________________________
    % 3. Compute new value of system (g1) and jacobian (D1) at m1
    [g1,D1,X1,Y1,~,~,~,~,Fevals]    = foc_con_2D(m1,h,Xmin,Xmax,Ymin,Ymax,c,param,glob);
    dist_m                          = max(abs(m1-m0));
    dist_x                          = max(abs([X0-X1;Y0-Y1]));
    %______________________________________________________________________
    % 4. Update
    g0          = real(g1);
    D0          = real(D1);
    m0          = real(m1);
    X0          = X1;
    Y0          = Y1;
    %______________________________________________________________________
    fprintf('%i. m = [%1.2f, %1.2f]\t[X,Y]=[%1.2f,%1.2f]\tsum(g) = %1.2e\tFevals = %2i\n',iter0,m0',X1,Y1,sum(g0),Fevals);
    flag1       = dist_x<1e-7;
    flag2       = dist_m<1e-7;
    flag3       = (iter0==max_iter);
    if flag2 || flag3;break;end
end
tocnewton = toc;
[FOC,D,X,Y,L1_X,L2_X,L1_Y,L2_Y,Fevals] = foc_con_2D(m0,h,Xmin,Xmax,Ymin,Ymax,c,param,glob);

%% Print results

fprintf('1. Fsolve\n');
fprintf('mstar      = %1.3f\t%1.3f\n',mstarfs');
fprintf('FOC(mstar) = %1.3f\t%1.3f\n',FOCfs');
fprintf('xstar      = %1.3f\t%1.3f\n',[Xfs,Yfs]);
fprintf('lambda1_X  = %1.3f\n',L1_Xfs);
fprintf('lambda2_X  = %1.3f\n',L2_Xfs);
fprintf('lambda1_Y  = %1.3f\n',L1_Yfs);
fprintf('lambda2_Y  = %1.3f\n',L2_Yfs);
fprintf('time       = %1.3f\n\n',tocfsolve);

fprintf('2. Newton\n');
fprintf('mstar      = %1.3f\t%1.3f\n',m0');
fprintf('FOC(mstar) = %1.3f\t%1.3f\n',FOC');
fprintf('xstar      = %1.3f\t%1.3f\n',[X,Y]);
fprintf('lambda1_X  = %1.3f\n',L1_X);
fprintf('lambda2_X  = %1.3f\n',L2_X);
fprintf('lambda1_Y  = %1.3f\n',L1_Y);
fprintf('lambda2_Y  = %1.3f\n',L2_Y);
fprintf('time       = %1.3f\n\n',tocnewton);

if flag1
    fprintf('Stopped because of x - Iterations = %4i\n',iter0);
end
if flag2
    fprintf('Stopped because of m - Iterations = %4i\n',iter0);
end
if flag3
    fprintf('Stopped because of max_iter - Iterations = %4i\n',iter0);
end

